

local BASE, VERSION, MINOR = 'LibBuilder', '1.0', 1
local MAJOR = BASE..'-'..VERSION


local libTable = LibStub:NewLibrary(MAJOR, MINOR)
if not libTable then return end


--	external upvalues
local _G = _G
local LibStub = _G.LibStub
local type = _G.type
local select = _G.select

--	internal upvalues
local onUpdate, newLib, getLib, closeLib, buildErrorMessage
local clientMajor, clientMinor, clientTable
local frame

--	constants
local defaultVersion = '1.0'	--	used if we supply the shell
local errorStrings = {
	SHELL = '%s: %s!', 
	BADVERSION = '%s requires a textual major and numeric minor version', 
	NOLIB = '%s called but no library is being built', 
	LEFTOPEN = '%s, Rev. %s was not closed properly',
}


--	Internal; uses frame updates to check if a library was left open.
function onUpdate()
	--	hide frame so we don't get another update unless another library 
	--	is created, and do it now in case we have to throw the error
	frame:Hide()

	if clientMajor then
		--	Close it so the error only happens once.
		--	We have to cache the version info because it'll get wiped.
		local major, minor = closeLib()
		
		local message = buildErrorMessage('LEFTOPEN', major, minor)
		error(message)
	end
end


---	Creates a private workspace for a new client library.
--	@name LibBuilder:NewLib
--	@param		register	Optional, defaults to true. If true, this function 
--							will attempt to create the library in LibStub.
--	@param		major		The major version of the library being built.
--	@param		minor		The minor version of the library being built.
--	@param		workspace	Optional; passing a value will cause LibBuilder 
--							to reuse it as the client library's workspace.
--	@return		workspace	The workspace for the new library.
--	@return		interface	The table returned from LibStub. Only returned 
--							if 'register' was true.
--	@return		oldMinor	Minor version of the previous instance of the 
--							library.  Only returned if 'register' was true.
function newLib(...)
	local register, major, minor, workspace
	local interface, oldMinor		--	only used if 'register' is not false
	
	--	shift arguments based on the type of the first
	local t = type(...)
	
	if t == 'boolean' then
		--	use first value
		register, major, minor, workspace = ...
	elseif t == nil then
		--	replace first value
		register, major, minor, workspace = true, select(2, ...)
	else
		--	prepend default
		register, major, minor, workspace = true, ...
	end
	
	--	check for collisions
	if clientMajor then
		--	close it properly instead of just 'shortcutting' by overwriting
		local oldMajor, oldMinor = closeLib()
		
		--	We can't use error() because we need to return values for the new lib.
		local message = buildErrorMessage('LEFTOPEN', oldMajor, oldMinor)
		geterrorhandler()(message)
	end
	
	--	check the supplied version info
	if type(major) ~= 'string' or not tonumber(minor) then
		local message = buildErrorMessage('BADVERSION', 'NewLib')
		error(message)
	end
	
	if register then
		interface, oldMinor = LibStub:NewLibrary(major, minor)
	end
	
	--	create the workspace even if LibStub said not 
	--	to create the client library, because the 
	--	client's files need to check.
	
	--	show the frame so we can use onUpdate
	--	to check if this library was left open.
	frame:Show()

	clientMajor = major
	clientMinor = minor
	
	--	FIXME: allow 'false' as a provided workspace?
	--	if the programmer gave us a funky workspace, that's their problem
	clientWorkspace = workspace or {}
	
	--	make the return show as a single value if register was false
	if register then
		return clientWorkspace, interface, oldMinor
	else
		return clientWorkspace
	end
end


---	Retrieves the version info and workspace for the current client library.
--	@name LibBuilder:GetLib()
--	@return major		Major version of the library.
--	@return minor		Minor version of the library.
--	@return workspace	The workspace for the library.
function getLib()
	if not clientMajor then
		local message = buildErrorMessage('NOLIB', 'GetLib')
		error(message)
	end
	
	return clientMajor, clientMinor, clientWorkspace
end


---	Closes the active library. This is used for tracking, 
--	to make sure two libraries don't collide.
--	This function returns the same values as GetLib, as a shortcut 
--	in case the client library needs to do last-minute initialization.
--	@name LibBuilder:CloseLib
--	@see GetLib
function closeLib()
	--	we don't want programmers calling this to 
	--	suppress errors before they open their lib.
	if not clientMajor then
		local message = buildErrorMessage('NOLIB', 'CloseLib')
		error(message)
	end
	
	--	the library was closed properly, so disable the OnUpdate frame
	frame:Hide()
	
	--	grab values before reset so we can return them
	local major, minor, workspace = clientMajor, clientMinor, clientWorkspace
	clientMajor, clientMinor, clientWorkspace = nil, nil, nil
	
	return major, minor, workspace
end


--	Internal; builds an error message.
--	@param	errorName	Name of the error
--	@param	...			List of arguments applicable to the error type.
--	@return message		The complete error string.
function buildErrorMessage(errorName, ...)
	local errorFormat = errorStrings[errorName]
	assert(errorFormat)
	
	local body = errorFormat:format(...)
	local message = errorStrings.SHELL:format(MAJOR, body)

	return message
end


--	Create a frame so we can throw a warning if 
--	a library is still open when OnUpdate fires.
--	Use OnUpdate instead of ADDON_LOADED because an error in the loading 
--	library would load Blizzard_DebugTools, confuse this library, and 
--	seemed to prevent the real error message from showing.
frame = CreateFrame('Frame')
frame:Hide()
frame:SetScript('OnUpdate', onUpdate)


libTable.NewLib  	= function(self, ...) return newLib(...) end
libTable.GetLib  	= function(self, ...) return getLib(...) end
libTable.CloseLib  	= function(self, ...) return closeLib(...) end


--	Internal; used for the library's __call metamethod.
--	If this library table is called with no arguments, it 
--	will call GetLib. If the first argument is 'close', 
--	CloseLib will be called with any subsequent arguments. 
--	Otherwise, NewLib will be passed with the arguments.
local function libcall(self, ...)
	local argCount = select('#', ...)
	
	if argCount == 0 then
		return self:GetLib()
	elseif ... == 'close' then
		return self:CloseLib(select(2, ...))
	else
		return self:NewLib(...)
	end
end
setmetatable(libTable, { __call = libcall })
_G[MAJOR] = libTable


--	create the shell if one doesn't already exist
local shell = _G[BASE]		--	we'll reuse this local after the if block
if not shell then
	shell = {}
	_G[BASE] = shell
	
	shell.version = VERSION
	
	--	Internal; if one argument is provided and is a valid index 
	--	in this table, the matching value will be returned. If 
	--	multiple arguments are provided and the first is a valid 
	--	index, the remaining arugments will be passed as a call 
	--	to the matching value. If the first argument is 'not' a 
	--	valid index, all arguments will be passed as a call to 
	--	the default library.
	local function shellcall(self, ...)
		if self[...] then
			if select('#', ...) == 1 then
				return self[...]
			else
				return self[...](select(2, ...))
			end
		else
			return self[defaultVersion](...)
		end
	end
	
	--	Internal; returns the default version if key is 
	--	an empty string, otherwise checks for the key in 
	--	the default library.  Note that this metamethod 
	--	will not be called if the specified key is present 
	--	in the shell.
	local function shellindex(self, key)
		if key == '' then
			return self[defaultVersion]
		else
			return self[defaultVersion][key]
		end
	end
	
	setmetatable(shell, { __call = shellcall, __index = shellindex })
end

--	add current version to the shell
shell[VERSION] = libTable
